import asyncio
from contextlib import contextmanager
import html
import json
from types import SimpleNamespace

from .hooks import _ctx, _url_provider
from .paths import Paths
from .node import node_flatten, SafeText


WORD_INIT_CHARS = frozenset(
    'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    'abcdefghijklmnopqrstuvwxyz'
    '-_'
)
WORD_CONT_CHARS = frozenset(
    'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    'abcdefghijklmnopqrstuvwxyz'
    '0123456789-_'
)


def _parse_filter(filter):
    __tracebackhide__ = True

    filters = [[]]
    index = 0

    def space():
        nonlocal index
        while index < len(filter) and filter[index].isspace():
            index += 1

    def word():
        nonlocal index
        if index >= len(filter) or filter[index] not in WORD_INIT_CHARS:
            raise AssertionError(f'{index}: expected a word')
        start = index
        index += 1
        while index < len(filter) and filter[index] in WORD_CONT_CHARS:
            index += 1
        return filter[start:index]

    def value():
        nonlocal index
        start = index
        depth = 0
        while True:
            if filter.startswith('"', index):
                index += 1
                while not filter.startswith('"', index):
                    if filter.startswith('\\"', index):
                        index += 2
                    elif index < len(filter):
                        index += 1
                    else:
                        index = start
                        raise AssertionError(f'{index}: expected a value')
                index += 1
            elif (
                filter.startswith('{', index) or
                filter.startswith('[', index)
            ):
                depth += 1
                index += 1
            elif (
                filter.startswith('}', index) or
                filter.startswith(']', index)
            ):
                depth -= 1
                index += 1
            elif depth == 0 and filter.startswith('true'):
                index += 4
            elif depth == 0 and filter.startswith('false'):
                index += 5
            elif depth == 0 and filter.startswith('null'):
                index += 4
            elif depth == 0 and index < len(filter) and (
                filter[index].isdigit() or
                filter[index] == '-'
            ):
                if filter[index] == '-':
                    index += 1
                if index >= len(filter) or not filter[index].isdigit():
                    index = start
                    raise AssertionError(f'{index}: expected a value')
                index += 1
                while index < len(filter) and filter[index].isdigit():
                    index += 1
                if filter.startswith('.', index):
                    index += 1
                    if index >= len(filter) or not filter[index].isdigit():
                        index = start
                        raise AssertionError(f'{index}: expected a value')
                    index += 1
                    while index < len(filter) and filter[index].isdigit():
                        index += 1
            elif depth != 0 and index < len(filter):
                index += 1
            else:
                index = start
                raise AssertionError(f'{index}: expected a value')

            if depth == 0:
                break

        try:
            return json.loads(filter[start:index])
        except ValueError:
            index = start
            raise AssertionError(f'{index}: expected a value')

    space()
    while index < len(filter):
        if filter[index] == ',':
            index += 1
            filters.append([])
            space()
            continue

        if filter[index] == '>':
            deep = False
            index += 1
            space()
        else:
            deep = True

        predicates = []
        star = False

        if index < len(filter):
            if filter[index] == '*':
                star = True
                index += 1
            elif filter[index] in WORD_INIT_CHARS:
                predicates.append(('tag', word()))

        while index < len(filter):
            if filter[index] == '#':
                index += 1
                predicates.append(('prop', 'id', word()))
            elif filter[index] == '.':
                index += 1
                predicates.append(('class', word()))
            elif filter[index] == '[':
                index += 1
                key = word()
                if filter.startswith('=', index):
                    index += 1
                    value = value()
                    predicates.append(('prop', key, value))
                else:
                    predicates.append(('has_prop', key))
                if not filter.startswith(']', index):
                    raise AssertionError(f'{index}: expected right bracket')
                index += 1
            elif filter[index] == ':':
                index += 1
                selector = word()
                args = []
                if filter.startswith('(', index):
                    index += 1
                    space()
                    while not filter.startswith(')', index):
                        args.append(value())
                        space()
                        if filter.startswith(')', index):
                            break
                        if not filter.startswith(',', index):
                            raise AssertionError(
                                f'{index}: expected right par or comma'
                            )
                        index += 1
                        space()
                    index += 1
                predicates.append(('selector', selector, *args))
            else:
                break

        if not predicates and not star:
            raise AssertionError(f'{index}: expected a predicate')

        filters[-1].append((tuple(predicates), deep))
        space()

    return tuple(map(tuple, filters))


def _children(node, *, deep=False):
    stack = []

    if isinstance(node[-1], tuple):
        stack.append((node, node_flatten((None, {}, {}, *node[-1][3:]))))

    while stack:
        path, nodes = stack[-1]
        try:
            node = next(nodes)
        except StopIteration:
            stack.pop()
            continue

        node = (*path, node)
        yield node

        if deep and isinstance(node[-1], tuple):
            stack.append((node, node_flatten((None, {}, {}, *node[-1][3:]))))


def _find(nodes, filter):
    for predicates, deep in filter:
        nodes = [
            child
            for node in nodes
            for child in _children(node, deep=deep)
        ]

        for predicate in predicates:
            if predicate[0] == 'tag':
                _, tag = predicate
                nodes = [node for node in nodes if (
                    isinstance(node[-1], tuple) and
                    node[-1][0] == tag
                )]

            elif predicate[0] == 'prop':
                _, key, value = predicate
                nodes = [node for node in nodes if (
                    isinstance(node[-1], tuple) and
                    key in node[-1][1] and
                    node[-1][1][key] == value
                )]

            elif predicate[0] == 'has_prop':
                _, key = predicate
                nodes = [node for node in nodes if (
                    isinstance(node[-1], tuple) and
                    key not in node[-1][1]
                )]

            elif predicate[0] == 'class':
                _, classname = predicate
                nodes = [node for node in nodes if (
                    isinstance(node[-1], tuple) and
                    classname in node[-1][1].get('class', '').split()
                )]

            elif predicate[0] == 'selector':
                _, selector, *args = predicate

                if selector == 'eq':
                    try:
                        index, = args
                    except ValueError:
                        raise AssertionError(
                            ':eq expects 1 argument'
                        ) from None
                    if not isinstance(index, int):
                        raise AssertionError(':eq index should be an int')
                    try:
                        nodes = [nodes[index]]
                    except IndexError:
                        raise AssertionError(
                            ':eq index out of range'
                        ) from None

                elif selector == 'text':
                    try:
                        content, = args
                    except ValueError:
                        raise AssertionError(
                            ':text expects 1 argument'
                        ) from None
                    if not isinstance(content, str):
                        raise AssertionError(':text content should be a str')
                    nodes = [node for node in nodes if (
                        _text([node]) == content
                    )]

                elif selector == 'contains':
                    try:
                        content, = args
                    except ValueError:
                        raise AssertionError(
                            ':text expects 1 argument'
                        ) from None
                    if not isinstance(content, str):
                        raise AssertionError(':text content should be a str')
                    nodes = [node for node in nodes if (
                        content in _text([node])
                    )]

                else:
                    raise AssertionError(f'unknown selector: {selector}')

            else:
                raise AssertionError(f'unknown predicate: {predicate[0]}')

    return nodes


def _text(nodes):
    parts = []
    safe = False

    for node in nodes:
        for child in _children(node, deep=True):
            if isinstance(child[-1], str):
                if safe:
                    parts.append(html.escape(child[-1]))
                else:
                    parts.append(child[-1])

            elif isinstance(child[-1], SafeText):
                if not safe:
                    parts = [html.escape(part) for part in parts]
                    safe = True
                parts.append(child[-1].text)

    text = ''.join(parts)
    if safe:
        text = SafeText(text)
    return text


NO_VALUE = object()


class Assertion:

    def __init__(self, session, actions):
        self._session = session
        self._actions = actions

    def _holds(self):
        actions = self._actions
        if not actions or actions[-1][0] == 'find':
            actions = (*actions, ('exists',))

        nodes = [(self._session._result,)]
        for action in actions:
            if action[0] == 'find':
                _, filters = action
                nodes = [
                    node
                    for filter in filters
                    for node in _find(nodes, filter)
                ]
            elif action[0] == 'parent':
                nodes = [
                    node[:-1]
                    for node in nodes
                    if len(node) > 1
                ]

            elif action[0] == 'exists':
                if not nodes:
                    return 'node does not exist'

            elif action[0] == 'not_exists':
                if nodes:
                    return 'node exists'

            elif action[0] == 'has_text':
                _, expected = action
                actual = _text(nodes)
                if actual != expected:
                    return f'node has text {actual!r} instead of {expected!r}'

            elif action[0] == 'not_has_text':
                _, expected = action
                if _text(nodes) == expected:
                    return f'node has text {expected!r}'

            elif action[0] == 'has_prop':
                _, key = action

                try:
                    path, = nodes
                except ValueError:
                    if nodes:
                        return 'cannot check props of multiple nodes'
                    else:
                        return 'no node to check props of'
                node = path[-1]

                if not isinstance(node, tuple) or key not in node[1]:
                    return f'node does not have prop {key}'

            elif action[0] == 'not_has_prop':
                _, key = action

                try:
                    path, = nodes
                except ValueError:
                    if nodes:
                        return 'cannot check props of multiple nodes'
                    else:
                        return 'no node to check props of'
                node = path[-1]

                if isinstance(node, tuple) and key in node[1]:
                    return f'node does have prop {key}'

            elif action[0] == 'prop':
                _, key, expected = action

                try:
                    path, = nodes
                except ValueError:
                    if nodes:
                        return 'cannot check props of multiple nodes'
                    else:
                        return 'no node to check props of'
                node = path[-1]

                try:
                    assert isinstance(node, tuple)
                    actual = node[1][key]
                except (AssertionError, KeyError):
                    return f'node does not have prop {key}'
                else:
                    if actual != expected:
                        return (
                            f'node prop {key} has value {actual!r} instead of '
                            f'{expected!r}'
                        )

            elif action[0] == 'not_prop':
                _, key, expected = action

                try:
                    path, = nodes
                except ValueError:
                    if nodes:
                        return 'cannot check props of multiple nodes'
                    else:
                        return 'no node to check props of'
                node = path[-1]

                try:
                    assert isinstance(node, tuple)
                    actual = node[1][key]
                except (AssertionError, KeyError):
                    pass
                else:
                    if actual == expected:
                        return f'node prop {key} does have value {expected!r}'

            elif action[0] == 'has_len':
                _, expected = action
                actual = len(nodes)
                if actual != expected:
                    return f'node has len {actual} instead of {expected}'

            elif action[0] == 'has_cookie':
                _, key = action
                if key not in self._session._cookies:
                    return f'node does not have cookie {key}'

            elif action[0] == 'not_has_cookie':
                _, key = action
                if key in self._session._cookies:
                    return f'session does have cookie {key}'

            elif action[0] == 'cookie':
                _, key, expected = action
                try:
                    actual = self._session._cookies[key]
                except KeyError:
                    return f'session does not have cookie {key}'
                else:
                    if actual != expected:
                        return (
                            f'session cookie {key} has value {actual!r} '
                            f'instead of {expected!r}'
                        )

            elif action[0] == 'not_cookie':
                _, key, expected = action
                try:
                    actual = self._session._cookies[key]
                except KeyError:
                    pass
                else:
                    if actual == expected:
                        return f'session cookie {key} has value {actual!r}'

            elif action[0] == 'url':
                _, expected = action
                if self._session._url != expected:
                    return (
                        f'session has url {self._session.url!r} instead of '
                        f'{expected!r}'
                    )

            elif action[0] == 'not_url':
                _, expected = action
                if self._session._url == expected:
                    return f'session has url {expected!r}'

            elif action[0] == 'click':
                try:
                    path, = nodes
                except ValueError:
                    if nodes:
                        return 'cannot click multiple nodes'
                    else:
                        return 'no node to click'
                node = path[-1]

                if (
                    not isinstance(node, tuple) or
                    'onclick' not in node[1] or
                    not callable(node[1]['onclick'])
                ):
                    return 'node is not clickable'

                loop = asyncio.get_running_loop()
                event = SimpleNamespace(type='click')
                loop.call_soon(node[1]['onclick'], event)

            elif action[0] == 'input':
                _, value = action

                try:
                    path, = nodes
                except ValueError:
                    if nodes:
                        return 'cannot input on multiple nodes'
                    else:
                        return 'no node to click'
                node = path[-1]

                if (
                    not isinstance(node, tuple) or
                    'oninput' not in node[1] or
                    not callable(node[1]['oninput'])
                ):
                    return 'node does not accept input'

                loop = asyncio.get_running_loop()
                event = SimpleNamespace(type='input', value=value)
                loop.call_soon(node[1]['oninput'], event)

            else:
                raise ValueError(f'unknown action: {action[0]}')

        return None

    async def _aholds(self, timeout):
        loop = asyncio.get_running_loop()

        timeout_fut = loop.create_task(asyncio.sleep(timeout))
        while True:
            reason = self._holds()
            if reason is None:
                timeout_fut.cancel()
                return None

            change_fut = loop.create_task(self._session._change.wait())
            await asyncio.wait(
                [change_fut, timeout_fut],
                return_when=asyncio.FIRST_COMPLETED,
            )
            if timeout_fut.done():
                if not change_fut.done():
                    change_fut.cancel()
                return reason

    def wait(self, timeout=None):
        if timeout is None:
            timeout = self._session._timeout

        loop = self._session._loop
        asyncio.set_event_loop(loop)
        try:
            reason = loop.run_until_complete(self._aholds(timeout))
        except AssertionError as e:
            reason = str(e)
        if reason is not None:
            __tracebackhide__ = True
            raise AssertionError(reason)
        self._session._update()

    def __bool__(self):
        __tracebackhide__ = True
        self.wait()
        return True

    def find(self, filter):
        __tracebackhide__ = True
        filters = _parse_filter(filter)
        return Assertion(self._session, (*self._actions, ('find', filters)))

    def parent(self):
        return Assertion(self._session, (*self._actions, ('parent',)))

    def exists(self):
        return Assertion(self._session, (*self._actions, ('exists',)))

    def not_exists(self):
        return Assertion(self._session, (*self._actions, ('not_exists',)))

    def has_text(self, text):
        return Assertion(self._session, (*self._actions, ('has_text', text)))

    def not_has_text(self, text):
        return Assertion(
            self._session,
            (*self._actions, ('not_has_text', text)),
        )

    def has_prop(self, key, value=NO_VALUE):
        if value is NO_VALUE:
            action = ('has_prop', key)
        else:
            action = ('prop', key, value)
        return Assertion(self._session, (*self._actions, action))

    def not_has_prop(self, key, value=NO_VALUE):
        if value is NO_VALUE:
            action = ('not_has_prop', key)
        else:
            action = ('not_prop', key, value)
        return Assertion(self._session, (*self._actions, action))

    def has_len(self, len):
        return Assertion(self._session, (*self._actions, ('has_len', len)))

    def has_cookie(self, key, value=NO_VALUE):
        if value is NO_VALUE:
            action = ('has_cookie', key)
        else:
            action = ('cookie', key, value)
        return Assertion(self._session, (*self._actions, action))

    def not_has_cookie(self, key, value=NO_VALUE):
        if value is NO_VALUE:
            action = ('not_has_cookie', key)
        else:
            action = ('not_cookie', key, value)
        return Assertion(self._session, (*self._actions, action))

    def has_url(self, url):
        return Assertion(self._session, (*self._actions, ('url', url)))

    def not_has_url(self, url):
        return Assertion(self._session, (*self._actions, ('not_url', url)))

    def click(self):
        return Assertion(self._session, (*self._actions, ('click',)))

    def input(self, value):
        return Assertion(self._session, (*self._actions, ('input', value)))


class TestSession(Assertion):

    __test__ = False

    def __init__(self, elem, *, url='/', cookies={}, timeout=3):
        super().__init__(self, ())

        self._url = url
        self._prev = []
        self._next = []
        self._cookies = cookies.copy()
        self._timeout = timeout
        self._elem = elem
        self._state, self._result = _url_provider(elem, value=url)._init()
        self._subscribers = set()

    def start(self, loop=None):
        if loop is None:
            loop = asyncio.new_event_loop()

        self._loop = loop
        self._run_task = loop.create_task(self._run())
        self._update()

    def stop(self):
        self._run_task.cancel()
        self._loop.close()

        del self._loop
        del self._run_task

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *exc):
        self.stop()

    async def _run(self):
        self._queue = asyncio.Queue()
        self._change = asyncio.Event()

        contexts = {}
        cookie_paths = {}

        def rerender_path(path):
            self._queue.put_nowait(('path', path))

        def push_url(url):
            self._queue.put_nowait(('push_url', url))

        def replace_url(url):
            self._queue.put_nowait(('replace_url', url))

        def set_cookie(key, value):
            self._queue.put_nowait(('set_cookie', key, value))

        def unset_cookie(key):
            self._queue.put_nowait(('unset_cookie', key))

        paths = Paths()

        while True:
            _ctx.static = False
            _ctx.rerender_path = rerender_path
            _ctx.push_url = push_url
            _ctx.replace_url = replace_url
            _ctx.set_cookie = set_cookie
            _ctx.unset_cookie = unset_cookie
            _ctx.contexts = contexts
            _ctx.cookies = self._cookies
            _ctx.cookie_paths = cookie_paths
            _ctx.rerender_paths = paths
            _ctx.path = []
            try:
                self._state, self._result = (
                    _url_provider(self._elem, value=self._url)
                    ._render(self._state, self._result)
                )
            finally:
                del _ctx.static
                del _ctx.rerender_path
                del _ctx.push_url
                del _ctx.replace_url
                del _ctx.set_cookie
                del _ctx.unset_cookie
                del _ctx.contexts
                del _ctx.cookies
                del _ctx.cookie_paths
                del _ctx.rerender_paths
                del _ctx.path

            self._change.set()
            self._change.clear()
            for callback in self._subscribers:
                callback()

            changes = [await self._queue.get()]
            while not self._queue.empty():
                changes.append(self._queue.get_nowait())

            paths = Paths()
            for change in changes:
                if change[0] == 'path':
                    _, path = change
                    paths[path] = None
                elif change[0] == 'prev_url':
                    self._next.append(self._url)
                    self._url = self._prev.pop()
                elif change[0] == 'next_url':
                    self._prev.append(self._url)
                    self._url = self._next.pop()
                elif change[0] in 'push_url':
                    self._prev.append(self._url)
                    _, self._url = change
                    self._next.clear()
                elif change[0] in 'replace_url':
                    _, self._url = change
                elif change[0] == 'set_cookie':
                    _, key, value = change
                    self._cookies[key] = value
                    for path in cookie_paths.get(key, []):
                        paths[path] = None
                elif change[0] == 'unset_cookie':
                    _, key = change
                    del self._cookies[key]
                    for path in cookie_paths.get(key, []):
                        paths[path] = None
                else:
                    raise ValueError(f'unknown change: {change[0]}')

    def prev(self):
        self._queue.put_nowait(('prev_url',))
        self._update()

    def next(self):
        self._queue.put_nowait(('next_url',))
        self._update()

    def _update(self):
        asyncio.set_event_loop(self._loop)
        while self._loop._ready:
            self._loop.stop()
            self._loop.run_forever()


@contextmanager
def run_together(self, *sessions, loop=None):
    if loop is None:
        loop = asyncio.new_event_loop()
        loop_owner = True
    else:
        loop_owner = False

    for session in sessions:
        session._loop = loop
        session._run_task = loop.create_task(session._run())

    try:
        asyncio.set_event_loop(loop)
        while loop._ready:
            loop.stop()
            loop.run_forever()

        yield sessions
    finally:
        for session in reversed(sessions):
            session._run_task.cancel()
            del session._loop
            del session._run_task
        if loop_owner:
            loop.close()



from collections import defaultdict, deque
import html
from itertools import islice
import json

from .events import CallbackWrapper


class SafeText:

    __slots__ = ['text']

    def __init__(self, text):
        self.text = text

    def __bool__(self):
        return bool(self.text)

    def __eq__(self, other):
        if isinstance(other, SafeText):
            return other.text == self.text
        elif isinstance(other, str):
            return html.escape(other, quote=False) == self.text
        else:
            return False

    def __hash__(self):
        return hash((SafeText, self.text))


def clean_value(value):
    if not callable(value):
        return value

    args = {
        'prevent_default': False,
        'stop_propagation': False,
        'stop_immediate_propagation': False,
    }

    while isinstance(value, CallbackWrapper):
        args[value.key] = value.value
        value = value.callback

    parts = ['call(event']
    for arg in [
        'prevent_default',
        'stop_propagation',
        'stop_immediate_propagation',
    ]:
        parts.append(', ')
        parts.append(json.dumps(args[arg]))
    parts.append(')')
    return ''.join(parts)


def clean_node(node):
    if not isinstance(node, tuple):
        return node

    tag, props, _, *children = node

    cleaned_props = {}
    for key, value in props.items():
        value = clean_value(value)
        if value is False:
            continue
        if value is True:
            value = ''
        if not isinstance(value, str):
            value = json.dumps(value)
        cleaned_props[key] = value

    cleaned_children = []
    stack = [iter(children)]
    while stack:
        try:
            child = next(stack[-1])
        except StopIteration:
            stack.pop()
            continue

        if isinstance(child, tuple) and child[0] is None:
            stack.append(islice(child, 3, None))
            continue

        if child is None:
            continue

        child = clean_node(child)
        if (
            cleaned_children and
            isinstance(child, str) and
            isinstance(cleaned_children[-1], str)
        ):
            cleaned_children[-1] += child
        else:
            cleaned_children.append(child)

    return (tag, cleaned_props, *cleaned_children)


def node_flatten(node):
    if not isinstance(node, tuple) or node[0] is not None:
        yield node
        return {(): 0}

    stack = [(enumerate(islice(node, 3, None)), ())]
    flat_index = 0
    path_indexes = {}

    while stack:
        nodes, path = stack[-1]
        try:
            index, node = next(nodes)
        except StopIteration:
            stack.pop()
            continue

        if isinstance(node, tuple) and node[0] is None:
            stack.append((enumerate(islice(node, 3, None)), (*path, index)))
            continue

        if isinstance(node, (str, SafeText)):
            while stack:
                nodes, path = stack[-1]
                try:
                    index, next_node = next(nodes)
                except StopIteration:
                    stack.pop()
                    continue

                if isinstance(next_node, tuple) and next_node[0] is None:
                    stack.append((
                        enumerate(islice(next_node, 3, None)),
                        (*path, index),
                    ))
                    continue

                if isinstance(next_node, str):
                    if isinstance(node, str):
                        node += next_node
                    else:
                        node = SafeText(
                            node.text +
                            html.escape(next_node, quote=False)
                        )
                elif isinstance(next_node, SafeText):
                    if isinstance(node, str):
                        node = SafeText(
                            html.escape(node, quote=False) +
                            next_node.text
                        )
                    else:
                        node = SafeText(node.text + next_node.text)
                else:
                    if node:
                        yield node
                        flat_index += 1
                    node = next_node
                    break

        if node is not None:
            yield node
            if not isinstance(node, (str, SafeText)):
                path_indexes[(*path, index)] = flat_index
            flat_index += 1

    return path_indexes


def node_get(node, path):
    for index in path:
        if node[0] is not None:
            node = (None, {}, {}, *node[3:])
        node = next(islice(node_flatten(node), index, None))
    return node


def node_parts(node):
    if isinstance(node, SafeText):
        yield node.text
        return

    if isinstance(node, str):
        yield html.escape(node, quote=False)
        return

    if node is None:
        return

    tag, props, _, *children = node

    if tag is not None:
        yield '<'
        yield tag
        for key, value in props.items():
            value = clean_value(value)
            if value is False:
                continue
            yield ' '
            yield key
            if value is True:
                continue
            yield '="'
            if not isinstance(value, str):
                value = json.dumps(value)
            yield html.escape(value)
            yield '"'
        yield '>'

    for child in children:
        yield from node_parts(child)

    if tag is not None:
        yield '</'
        yield tag
        yield '>'


def node_diff(old_node, new_node, path=()):
    old_nodes_iter = node_flatten(old_node)
    old_nodes = []
    while True:
        try:
            old_nodes.append(next(old_nodes_iter))
        except StopIteration as e:
            old_path_indexes = e.value
            break

    new_nodes_iter = node_flatten(new_node)
    new_nodes = []
    while True:
        try:
            new_nodes.append(next(new_nodes_iter))
        except StopIteration as e:
            new_path_indexes = e.value
            break

    index_mapping = {}
    for new_path, new_index in new_path_indexes.items():
        node = new_node
        old_path = []
        for index in new_path:
            try:
                old_index = node[2][index]
            except KeyError:
                break
            node = node[index + 3]
            old_path.append(old_index)
        else:
            try:
                old_index = old_path_indexes[tuple(old_path)]
            except KeyError:
                pass
            else:
                index_mapping[new_index] = old_index

    old_str_indexes = defaultdict(deque)

    for old_index, node in enumerate(old_nodes):
        if isinstance(node, (str, SafeText)):
            old_str_indexes[node].append(old_index)

    for new_index, node in enumerate(new_nodes):
        if isinstance(node, (str, SafeText)):
            try:
                old_index = old_str_indexes[node].popleft()
            except IndexError:
                pass
            else:
                index_mapping[new_index] = old_index

    rev_index_mapping = {value: key for key, value in index_mapping.items()}

    inserts = deque()
    removes = 0
    waiting = {}
    index = 0
    old_index = 0

    for new_index, new_node in enumerate(new_nodes):
        if new_index not in index_mapping:
            inserts.append(new_node)
            continue

        try:
            curr_index = waiting.pop(new_index)
        except KeyError:
            target_old_index = index_mapping[new_index]
            while old_index < target_old_index:
                try:
                    new_index_ = rev_index_mapping[old_index]
                except KeyError:
                    removes += 1
                else:
                    waiting[new_index] = index
                    index += 1
                old_index += 1
        else:
            index -= 1
            yield ('move', *path, curr_index, index)
            for new_index_, curr_index_ in waiting.items():
                if curr_index < curr_index_:
                    waiting[new_index_] = curr_index_ - 1

        while inserts:
            if removes:
                action = 'replace'
                removes -= 1
            else:
                action = 'insert'
            yield (action, *path, index, clean_node(inserts.popleft()))
            index += 1

        while removes:
            yield ('remove', *path, index)
            removes -= 1

        old_node = old_nodes[old_index]

        if old_node is new_node:
            pass
        elif isinstance(old_node, tuple) and isinstance(new_node, tuple):
            _, old_props, old_mapping, *old_children = old_node
            _, new_props, new_mapping, *new_children = new_node

            for key in set(old_props) - set(new_props):
                yield ('unset', *path, index, key)

            for key, value in new_props.items():
                value = clean_value(value)
                if (
                    key not in old_props or
                    clean_value(old_props[key]) != value
                ):
                    if value is False:
                        yield ('unset', *path, index, key)
                        continue
                    if value is True:
                        value = ''
                    yield ('set', *path, index, key, value)

            yield from node_diff(
                (None, {}, old_mapping, *old_children),
                (None, {}, new_mapping, *new_children),
                (*path, index),
            )
        elif old_node != new_node:
            yield ('replace', *path, index, clean_node(new_node))

        old_index += 1
        index += 1

    removes += len(old_nodes) - old_index

    while inserts:
        if removes:
            action = 'replace'
            removes -= 1
        else:
            action = 'insert'
        yield (action, *path, index, clean_node(inserts.popleft()))
        index += 1

    while removes:
        yield ('remove', *path, index)
        removes -= 1



import asyncio
from types import SimpleNamespace

from .context import create_context


_ctx = SimpleNamespace()


def use_ref(**kwargs):
    if isinstance(_ctx.refs, list):
        ref = SimpleNamespace(**kwargs)
        _ctx.refs.append(ref)
    else:
        try:
            ref = next(_ctx.refs)
        except StopIteration:
            raise ValueError('more refs used than previous render') from None
    return ref


def use_state(initial_value):
    ref = use_ref()
    _ctx.static = False
    ref.path = tuple(_ctx.path)
    ref.rerender_path = _ctx.rerender_path

    if not hasattr(ref, 'value'):
        if callable(initial_value):
            initial_value = initial_value()

        def set_value(value):
            if callable(value):
                value = value(ref.value)

            ref.value = value
            ref.rerender_path(ref.path)

        ref.value = initial_value
        ref.set_value = set_value

    return ref.value, ref.set_value


def use_memo(*key):
    def decorator(callback):
        ref = use_ref()
        if not hasattr(ref, 'key') or ref.key != key:
            ref.key = key
            ref.value = callback()
        return ref.value
    return decorator


def use_callback(*key):
    def decorator(callback):
        return use_memo(*key)(lambda: callback)
    return decorator


def use_effect(*key):
    def decorator(callback):
        ref = use_ref()
        if not hasattr(ref, 'key') or ref.key != key:
            loop = asyncio.get_running_loop()

            ref.key = key

            if hasattr(ref, '_vivi_cleanup'):
                ref._vivi_cleanup()
                del ref._vivi_cleanup

            @loop.call_soon
            def wrapped_callback():
                cleanup = callback()
                if callable(cleanup):
                    ref._vivi_cleanup = lambda: loop.call_soon(cleanup)

    return decorator


_url_provider, use_url = create_context()


def use_push_url():
    _ctx.static = False
    return _ctx.push_url


def use_replace_url():
    _ctx.static = False
    return _ctx.replace_url


def use_future(fut, sentinel=None):
    ref = use_ref(fut=None)

    ref.path = tuple(_ctx.path)
    ref.rerender_path = _ctx.rerender_path

    if fut is not ref.fut:
        ref.fut = fut
        if fut is not None and not fut.done():
            @fut.add_done_callback
            def on_fut_done(fut):
                if fut is ref.fut:
                    ref.rerender_path(ref.path)

    if fut is not None and fut.done():
        return fut.result()
    else:
        return sentinel


NO_DEFAULT = object()


def use_cookie(key, default=NO_DEFAULT):
    ref = use_ref()

    if not hasattr(ref, 'key') or ref.key != key:
        ref.key = key

        if hasattr(ref, '_vivi_cleanup'):
            ref._vivi_cleanup()

        cookie_paths = _ctx.cookie_paths
        path = tuple(_ctx.path)

        cookie_paths.setdefault(key, set()).add(path)

        def cleanup():
            paths = cookie_paths[key]
            paths.remove(path)
            if not paths:
                del cookie_paths[key]

        ref._vivi_cleanup = cleanup

    try:
        return _ctx.cookies[key]
    except KeyError:
        if default is NO_DEFAULT:
            raise
        else:
            return default


def use_set_cookie():
    _ctx.static = False
    return _ctx.set_cookie


def use_unset_cookie():
    _ctx.static = False
    return _ctx.unset_cookie



from .paths import Paths


def create_context(initial_value=None):
    from .elements import component, fragment
    from .hooks import _ctx, use_ref

    key = object()

    def get_context():
        try:
            res = _ctx.contexts[key]
        except KeyError:
            if callable(initial_value):
                value = initial_value()
            else:
                value = initial_value
            res = (value, Paths(), Paths())
            _ctx.contexts[key] = res
        return res

    @component
    def context_provider(value, children=()):
        initial_value, providers, receivers = get_context()
        path = tuple(_ctx.path)

        if path not in providers or providers[path] is not value:
            providers[path] = value
            _ctx.rerender_paths.update(
                (path, None)
                for path in receivers.children(path, providers)
            )

        ref = use_ref()
        if not hasattr('ref', '_vivi_cleanup'):
            def cleanup():
                del providers[path]

            ref._vivi_cleanup = cleanup

        return fragment(*children)

    def use_context():
        initial_value, providers, receivers = get_context()
        path = tuple(_ctx.path)

        ref = use_ref()
        if not hasattr(ref, '_vivi_cleanup'):
            receivers[path] = None

            def cleanup():
                del receivers[path]

            ref._vivi_cleanup = cleanup

        try:
            parent = providers.closest_parent(path)
        except KeyError:
            return initial_value
        else:
            return providers[parent]

    return context_provider, use_context



class CallbackWrapper:

    def __init__(self, callback, key, value):
        self.callback = callback
        self.key = key
        self.value = value

    def __eq__(self, other):
        return (
            isinstance(other, CallbackWrapper) and
            other.callback == self.callback and
            other.key == self.key and
            other.value == self.value
        )

    def __call__(self, *args, **kwargs):
        return self.callback(*args, **kwargs)


def prevent_default(callback):
    return CallbackWrapper(callback, 'prevent_default', True)


def stop_propagation(callback):
    return CallbackWrapper(callback, 'stop_propagation', True)


def stop_immediate_propagation(callback):
    return CallbackWrapper(callback, 'stop_immediate_propagation', True)



import asyncio
from itertools import islice
import json
from pathlib import Path
from types import SimpleNamespace
from uuid import uuid4

from starlette.applications import Starlette
from starlette.routing import Route, WebSocketRoute, Mount
from starlette.staticfiles import StaticFiles
from starlette.responses import Response
from starlette.websockets import WebSocketDisconnect

from .hooks import _ctx, _url_provider
from .node import SafeText, node_get, node_parts, node_diff
from .paths import Paths


DOCTYPE = SafeText('<!doctype html>')
SCRIPT_BEFORE, SCRIPT_AFTER = (
    Path(__file__).parent.joinpath('app.js')
    .read_text().split('{{socket_url}}')
)


def wrap(result, script=None, prev_head=None):
    while isinstance(result, tuple) and result[0] is None and len(result) == 4:
        result = result[3]

    if not isinstance(result, tuple) or result[0] != 'html':
        result = ('html', {}, {0: 0}, ('body', {}, {0: 0}, result))

    head = None
    body = None

    stack = [islice(result, 3, None)]
    while stack:
        try:
            node = next(stack[-1])
        except StopIteration:
            stack.pop()
            continue

        if isinstance(node, tuple) and node[0] is None:
            stack.append(islice(result, 3, None))
        elif isinstance(node, tuple) and node[0] == 'head':
            assert head is None
            head = node
        elif isinstance(node, tuple) and node[0] == 'body':
            assert body is None
            body = node
        else:
            raise ValueError(f'unexpected node in html: {node}')

    original_head = head

    if script is not None:
        if head is None:
            head = ('head', {}, {0: 0}, script)
        else:
            _, props, mapping, *children = head
            if head is prev_head:
                mapping = {i: i for i in range(len(children))}
            new_mapping = {0: 0}
            for index, prev_index in mapping.items():
                new_mapping[index + 1] = prev_index + 1
            head = ('head', {}, new_mapping, script, *children)

    if head is None:
        head = ('head', {}, {})
    if body is None:
        body = ('body', {}, {})

    result = (
        None, {}, {0: 0, 1: 1},
        DOCTYPE,
        ('html', result[1], {0: 0, 1: 1}, head, body),
    )
    return result, original_head


class Vivi(Starlette):

    def __init__(
        self, elem, *,
        debug=False,
        static_path=None,
        static_route='/static',
        on_startup=[],
        on_shutdown=[],
    ):
        routes = []

        if static_path is not None:
            routes.append(Mount(
                static_route,
                app=StaticFiles(directory=static_path),
            ))

        routes.append(Route(
            '/{path:path}',
            endpoint=self._http,
            methods=['GET'],
            name='http',
        ))
        routes.append(WebSocketRoute(
            '/{session_id:uuid}',
            endpoint=self._websocket,
            name='websocket',
        ))

        super().__init__(
            debug=debug,
            routes=routes,
            on_startup=on_startup,
            on_shutdown=on_shutdown,
        )

        self._elem = elem
        self._sessions = {}

    async def _http(self, request):
        queue = asyncio.Queue()
        url = request['path']
        contexts = {}
        cookies = request.cookies
        cookie_paths = {}

        def rerender_path(path):
            queue.put_nowait(('path', path))

        def push_url(url):
            queue.put_nowait(('push_url', url))

        def replace_url(url):
            queue.put_nowait(('replace_url', url))

        def set_cookie(key, value):
            queue.put_nowait(('set_cookie', key, value))

        def unset_cookie(key):
            queue.put_nowait(('unset_cookie', key))

        elem = _url_provider(self._elem, value=url)

        _ctx.static = True
        _ctx.rerender_path = rerender_path
        _ctx.push_url = push_url
        _ctx.replace_url = replace_url
        _ctx.set_cookie = set_cookie
        _ctx.unset_cookie = unset_cookie
        _ctx.contexts = contexts
        _ctx.cookies = cookies
        _ctx.cookie_paths = cookie_paths
        _ctx.rerender_paths = Paths()
        _ctx.path = []
        try:
            state, result = elem._render(*elem._init())
            static = _ctx.static
        finally:
            del _ctx.static
            del _ctx.rerender_path
            del _ctx.push_url
            del _ctx.replace_url
            del _ctx.set_cookie
            del _ctx.unset_cookie
            del _ctx.contexts
            del _ctx.cookies
            del _ctx.cookie_paths
            del _ctx.rerender_paths
            del _ctx.path

        if static:
            self._elem._unmount(state, result)
            result, _ = wrap(result)
        else:
            session_id = uuid4()
            script = ('script', {}, {0: 0}, SafeText(
                SCRIPT_BEFORE +
                json.dumps(
                    request.url_for('websocket', session_id=session_id)
                ) +
                SCRIPT_AFTER
            ))
            base_result = result
            result, head = wrap(result, script)

            self._sessions[session_id] = (
                state,
                base_result,
                head,
                script,
                queue,
                rerender_path,
                push_url,
                replace_url,
                set_cookie,
                unset_cookie,
                url,
                contexts,
                cookies,
                cookie_paths,
            )

            def session_timeout():
                try:
                    state, result, *_ = self._sessions.pop(session_id)
                except KeyError:
                    return
                elem._unmount(state, result)

            loop = asyncio.get_running_loop()
            loop.call_later(5, session_timeout)

        return Response(
            ''.join(node_parts(result)),
            media_type='text/html',
            headers={'connection': 'keep-alive'},
        )

    async def _websocket(self, socket):
        loop = asyncio.get_running_loop()

        session_id = socket.path_params['session_id']
        try:
            (
                state,
                result,
                head,
                script,
                queue,
                rerender_path,
                push_url,
                replace_url,
                set_cookie,
                unset_cookie,
                url,
                contexts,
                cookies,
                cookie_paths,
            ) = self._sessions.pop(session_id)
        except KeyError:
            await socket.close()
            return

        await socket.accept()

        receive_fut = loop.create_task(socket.receive_json())
        queue_fut = loop.create_task(queue.get())

        while True:
            await asyncio.wait(
                [receive_fut, queue_fut],
                return_when=asyncio.FIRST_COMPLETED,
            )

            if receive_fut.done():
                try:
                    event_type, *path, details = receive_fut.result()
                except WebSocketDisconnect:
                    elem = _url_provider(self._elem, value=url)
                    wrapped_result, _ = wrap(result, script, head)
                    elem._unmount(state, wrapped_result)
                    return

                if event_type == 'pop_url':
                    assert not path
                    queue.put_nowait(('pop_url', details))
                else:
                    target = node_get(wrap(result, script)[0], path)
                    handler = target[1][f'on{event_type}']

                    event = SimpleNamespace(type=event_type, **details)
                    loop.call_soon(handler, event)

                receive_fut = loop.create_task(socket.receive_json())

            elif queue_fut.done():
                changes = [queue_fut.result()]
                while not queue.empty():
                    changes.append(queue.get_nowait())

                actions = []
                paths = Paths()
                for change in changes:
                    if change[0] == 'path':
                        _, path = change
                        paths[path] = None
                    elif change[0] in ('pop_url', 'push_url', 'replace_url'):
                        change_type, url = change
                        if change_type != 'pop_url':
                            actions.append(change)
                    elif change[0] == 'set_cookie':
                        _, key, value = change
                        cookies[key] = value
                        for path in cookie_paths.get(key, []):
                            paths[path] = None
                        actions.append(change)
                    elif change[0] == 'unset_cookie':
                        _, key = change
                        del cookies[key]
                        for path in cookie_paths.get(key, []):
                            paths[path] = None
                        actions.append(change)
                    else:
                        raise ValueError(f'unknown change: {change[0]}')

                old_result, _ = wrap(result, script)

                elem = _url_provider(self._elem, value=url)

                _ctx.static = False
                _ctx.rerender_path = rerender_path
                _ctx.push_url = push_url
                _ctx.replace_url = replace_url
                _ctx.set_cookie = set_cookie
                _ctx.unset_cookie = unset_cookie
                _ctx.contexts = contexts
                _ctx.cookies = cookies
                _ctx.cookie_paths = cookie_paths
                _ctx.rerender_paths = paths
                _ctx.path = []
                try:
                    state, result = elem._render(state, result)
                finally:
                    del _ctx.static
                    del _ctx.rerender_path
                    del _ctx.push_url
                    del _ctx.replace_url
                    del _ctx.set_cookie
                    del _ctx.unset_cookie
                    del _ctx.contexts
                    del _ctx.cookies
                    del _ctx.cookie_paths
                    del _ctx.rerender_paths
                    del _ctx.path

                new_result, head = wrap(result, script, head)

                actions.extend(node_diff(old_result, new_result))

                if actions:
                    await socket.send_text(json.dumps(
                        actions,
                        separators=(',', ':'),
                    ))

                queue_fut = loop.create_task(queue.get())



from collections.abc import MutableMapping


VALUE = object()


class Paths(MutableMapping):

    def __init__(self):
        self._data = {}

    def __getitem__(self, path):
        data = self._data
        for key in path:
            data = data[key]
        return data[VALUE]

    def __setitem__(self, path, value):
        data = self._data
        for key in path:
            data = data.setdefault(key, {})
        data[VALUE] = value

    def __delitem__(self, path):
        data = self._data
        parents = []
        for key in path:
            parents.append((data, key))
            data = data[key]
        del data[VALUE]
        while not data and parents:
            data, key = parents.pop()
            del data[key]

    def __iter__(self):
        todo = [(self._data, ())]
        while todo:
            data, path = todo.pop()
            for key, value in data.items():
                if key is VALUE:
                    yield path
                else:
                    todo.append((value, (*path, key)))

    def __len__(self):
        todo = [(self._data, ())]
        res = 0
        while todo:
            data, path = todo.pop()
            for key, value in data.items():
                if key is VALUE:
                    res += 1
                else:
                    todo.append((value, (*path, key)))
        return res

    def closest_parent(self, path):
        data = self._data

        if VALUE in data:
            res = 0
        else:
            res = None

        for index, key in enumerate(path):
            try:
                data = data[key]
            except KeyError:
                break
            else:
                if VALUE in data:
                    res = index + 1

        if res is None:
            raise KeyError

        return path[:res]

    def children(self, path, stop=set(), stop_at_value=False):
        data = self._data
        for key in path:
            try:
                data = data[key]
            except KeyError:
                return

        root = path
        todo = [(data, path)]
        while todo:
            data, path = todo.pop()

            if path is not root and VALUE in data:
                yield path
                if stop_at_value:
                    continue

            for key, value in data.items():
                if key is not VALUE:
                    subpath = (*path, key)
                    if subpath not in stop:
                        todo.append((value, subpath))

    def prune(self, path):
        data = self._data
        parents = []
        for key in path:
            parents.append((data, key))
            try:
                data = data[key]
            except KeyError:
                return
        data.clear()
        while not data and parents:
            data, key = parents.pop()
            del data[key]



from abc import ABC, abstractmethod

from .hooks import _ctx
from .node import SafeText


INCOMPATIBLE = object()
COMPATIBLE = object()
EQUIVALENT = object()

INDEX_KEY = object()


class Element(ABC):

    def __init__(self, props, children):
        try:
            self._key = props.pop('key')
        except KeyError:
            pass

        self._props = props
        self._children = children

    @abstractmethod
    def _copy(self, props, children):
        raise NotImplementedError

    @abstractmethod
    def _comp(self, elem):
        raise NotImplementedError

    @abstractmethod
    def _init(self):
        raise NotImplementedError

    @abstractmethod
    def _render(self, prev_state, prev_result):
        raise NotImplementedError

    @abstractmethod
    def _unmount(self, state, result):
        raise NotImplementedError

    @abstractmethod
    def _extract(self, state, result, key):
        raise NotImplementedError

    @abstractmethod
    def _insert(self, state, result, key, child_state, child_result):
        raise NotImplementedError

    def _clean_elem(self, elem):
        if elem is None or isinstance(elem, (SafeText, str)):
            elem = Literal(elem)

        if not isinstance(elem, Element) and not isinstance(elem, str):
            try:
                children = tuple(elem)
            except TypeError:
                pass
            else:
                elem = HTMLElement(None, {}, children)

        if not isinstance(elem, Element):
            elem = Literal(str(elem))

        return elem

    def _rerender(self, prev_elem, prev_state, prev_result):
        comp = self._comp(prev_elem)

        if comp is INCOMPATIBLE:
            prev_elem._unmount(prev_state, prev_result)
            _ctx.rerender_paths.prune(_ctx.path)
            prev_state, prev_result = self._init()

        if comp is EQUIVALENT and _ctx.path not in _ctx.rerender_paths:
            state = prev_state
            result = prev_result

            for subpath in _ctx.rerender_paths.children(
                _ctx.path, stop_at_value=True,
            ):
                path = _ctx.path
                _ctx.path = list(subpath)
                try:
                    state, result = self._rerender_path(
                        subpath[len(path):], state, result,
                    )
                finally:
                    _ctx.path = path
        else:
            state, result = self._render(prev_state, prev_result)

        return state, result

    def _rerender_path(self, path, state, result):
        stack = []
        elem = self

        for key in path:
            stack.append((elem, state, result, key))
            elem, state, result = elem._extract(state, result, key)

        state, result = elem._render(state, result)

        while stack:
            elem, prev_state, prev_result, key = stack.pop()
            state, result = elem._insert(
                prev_state, prev_result, key, state, result,
            )

        assert elem is self
        return state, result

    def __call__(self, *args, **kwargs):
        if 'children' in kwargs:
            raise ValueError('\'children\' is not allowed as a property name')

        return self._copy({**self._props, **kwargs}, (*self._children, *args))


class Literal(Element):

    def __init__(self, value):
        self._value = value

    def _copy(self, props, children):
        raise ValueError('literals cannot be copied')

    def _comp(self, elem):
        if not isinstance(elem, Literal):
            return INCOMPATIBLE
        elif elem._value != self._value:
            return COMPATIBLE
        else:
            return EQUIVALENT

    def _init(self):
        return None, self._value

    def _render(self, prev_state, prev_result):
        return None, self._value

    def _unmount(self, state, result):
        pass

    def _extract(self, state, result, key):
        raise ValueError('literals do not have children')

    def _insert(self, state, result, key, child_state, child_result):
        raise ValueError('literals do not have children')


class HTMLElement(Element):

    def __init__(self, tag, props, children):
        if tag is None and props:
            raise ValueError('fragment cannot have props')

        super().__init__(props, children)
        self._tag = tag

    def _copy(self, props, children):
        return HTMLElement(self._tag, props, children)

    def __eq__(self, other):
        return (
            isinstance(other, HTMLElement) and
            other._tag == self._tag and
            other._props == self._props and
            other._children == self._children
        )

    def _comp(self, elem):
        if elem == self:
            return EQUIVALENT
        elif isinstance(elem, HTMLElement) and elem._tag == self._tag:
            return COMPATIBLE
        else:
            return INCOMPATIBLE

    def _init(self):
        return {}, (self._tag, self._props, {})

    def _render(self, prev_state, prev_result):
        state = {}
        child_results = []
        child_prev_indexes = {}

        prev_state = prev_state.copy()

        for index, child in enumerate(map(self._clean_elem, self._children)):
            try:
                key = child._key
            except AttributeError:
                key = (INDEX_KEY, index)

            if key in state:
                raise ValueError('duplicate keys')

            try:
                prev_child, prev_child_state, prev_index = prev_state.pop(key)
            except KeyError:
                prev_child = Literal(None)
                prev_child_state = None
                prev_child_result = None
            else:
                prev_child_result = prev_result[prev_index + 3]
                child_prev_indexes[index] = prev_index

            _ctx.path.append(key)
            try:
                child_state, child_result = child._rerender(
                    prev_child, prev_child_state, prev_child_result,
                )
            finally:
                _ctx.path.pop()

            state[key] = (child, child_state, index)
            child_results.append(child_result)

        for prev_child, prev_child_state, prev_index in prev_state.values():
            prev_child_result = prev_result[prev_index + 3]
            prev_child._unmount(prev_child_state, prev_child_result)

        result = (self._tag, self._props, child_prev_indexes, *child_results)
        return state, result

    def _unmount(self, state, result):
        for child, child_state, index in state.values():
            child_result = result[index + 3]
            child._unmount(child_state, child_result)

    def _extract(self, state, result, key):
        child, child_state, index = state[key]
        child_result = result[index + 3]
        return child, child_state, child_result

    def _insert(self, state, result, key, child_state, child_result):
        child, _, index = state[key]
        state = {**state, key: (child, child_state, index)}

        tag, props, _, *children = result
        result = (
            tag, props,
            {i: i for i in range(len(children))},
            *children[:index], child_result, *children[index + 1:],
        )

        return state, result


class Component(Element):

    def __init__(self, func, props, children):
        super().__init__(props, children)
        self._func = func

    def _copy(self, props, children):
        return Component(self._func, props, children)

    def __eq__(self, other):
        return (
            isinstance(other, Component) and
            other._func == self._func and
            other._props == self._props and
            other._children == self._children
        )

    def _comp(self, elem):
        if elem == self:
            return EQUIVALENT
        elif isinstance(elem, Component) and elem._func == self._func:
            return COMPATIBLE
        else:
            return INCOMPATIBLE

    def _init(self):
        return (None, Literal(None), None), None

    def _render(self, prev_state, prev_result):
        refs, prev_elem, prev_elem_state = prev_state

        _ctx.refs = [] if refs is None else iter(refs)
        try:
            props = self._props
            if self._children:
                props = {**props, 'children': self._children}
            elem = self._clean_elem(self._func(**props))

            if refs is None:
                refs = tuple(_ctx.refs)
            else:
                try:
                    next(_ctx.refs)
                except StopIteration:
                    pass
                else:
                    raise ValueError('less refs used than previous render')
        finally:
            del _ctx.refs

        _ctx.path.append('render')
        try:
            elem_state, result = elem._rerender(
                prev_elem, prev_elem_state, prev_result,
            )
        finally:
            _ctx.path.pop()

        return (refs, elem, elem_state), result

    def _unmount(self, state, result):
        refs, elem, elem_state = state
        if isinstance(elem, Element):
            elem._unmount(elem_state, result)
        for ref in refs:
            if hasattr(ref, '_vivi_cleanup'):
                ref._vivi_cleanup()

    def _extract(self, state, result, key):
        assert key == 'render'
        _, child, child_state = state
        return child, child_state, result

    def _insert(self, state, result, key, child_state, child_result):
        assert key == 'render'
        refs, child, _ = state
        return (refs, child, child_state), child_result


class HTMLFactory:

    def __getattr__(self, name):
        return HTMLElement(name, {}, ())


h = HTMLFactory()


def component(func):
    return Component(func, {}, ())


fragment = HTMLElement(None, {}, ())



from starlette.routing import compile_path

from .elements import component, h, fragment, Component
from .hooks import use_url, use_push_url, use_callback, use_ref
from .events import prevent_default


@component
def link(to, children, add_active=False, **props):
    url = use_url()
    push_url = use_push_url()

    current = url == to

    @use_callback(push_url, to, current)
    @prevent_default
    def onclick(e):
        if not current:
            push_url(to)

    if add_active and url == to:
        try:
            props['class'] = f'{props["class"]} active'
        except KeyError:
            props['class'] = 'active'

    return h.a(href=to, onclick=onclick, **props)(*children)


@component
def router(children=(), not_found=fragment):
    ref = use_ref(routes=[], url=None, index=None, values=None)
    url = use_url()

    first_change = None

    if url != ref.url:
        ref.url = url
        ref.index = None
        first_change = 0

    for index, (pattern, elem) in enumerate(children):
        try:
            prev_pattern, prev_regex, prev_converters, _ = ref.routes[index]
        except IndexError:
            if first_change is None:
                first_change = index
            regex, _, converters = compile_path(pattern)
            ref.routes.append((pattern, regex, converters, elem))
        else:
            if pattern == prev_pattern:
                regex = prev_regex
                converters = prev_converters
            else:
                if first_change is None:
                    first_change = index
                regex, _, converters = compile_path(pattern)
            ref.routes[index] = (pattern, regex, converters, elem)

    if len(ref.routes) > len(children):
        if first_change is None:
            first_change = len(children)
        ref.routes[len(children):] = []

    if (
        first_change is not None and
        (ref.index is None or ref.index >= first_change)
    ):
        for index in range(first_change, len(ref.routes)):
            _, regex, converters, elem = ref.routes[index]
            match = regex.match(url)
            if match is not None:
                ref.index = index
                ref.values = {
                    key: converters[key].convert(value)
                    for key, value in match.groupdict()
                }
                break
        else:
            ref.index = None
            ref.values = None

    if ref.index is None:
        return not_found
    else:
        elem = ref.routes[ref.index][3]
        if isinstance(elem, Component):
            elem = elem(**ref.values)
        return elem
